from .django19 import ReverseOneToOneDescriptor, ForwardManyToOneDescriptor
from .django110 import MiddlewareMixin
from .fluent_compat import add_preserved_filters, smart_text, truncate_name, AUTH_USER_MODEL, get_models, get_model, get_app_label, get_app_names, get_user_model, is_queryset_empty, get_current_site, get_meta_model_name, is_installed

__all__ = (
    'get_user_model', 'AUTH_USER_MODEL',
    'transaction_atomic',
    'add_preserved_filters',
    'is_queryset_empty',
    'is_installed', 'get_models', 'get_model', 'get_app_names', 'get_app_label', 'truncate_name', 'get_meta_model_name', 'get_current_site',
    'ReverseOneToOneDescriptor', 'ForwardManyToOneDescriptor',
    'MiddlewareMixin',
    'smart_text',
)



from __future__ import absolute_import

import importlib
import sys
import traceback

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

__all__ = (
    'import_settings_class',
    'import_class',
    'import_apps_submodule',
)


def import_settings_class(setting_name):
    """
    Return the class pointed to be an app setting variable.
    """
    config_value = getattr(settings, setting_name)
    if config_value is None:
        raise ImproperlyConfigured("Required setting not found: {0}".format(setting_name))

    return import_class(config_value, setting_name)


def import_class(import_path, setting_name=None):
    """
    Import a class by name.
    """
    mod_name, class_name = import_path.rsplit('.', 1)

    # import module
    mod = import_module_or_none(mod_name)
    if mod is not None:
        # Loaded module, get attribute
        try:
            return getattr(mod, class_name)
        except AttributeError:
            pass

    # For ImportError and AttributeError, raise the same exception.
    if setting_name:
        raise ImproperlyConfigured("{0} does not point to an existing class: {1}".format(setting_name, import_path))
    else:
        raise ImproperlyConfigured("Class not found: {0}".format(import_path))


def import_apps_submodule(submodule):
    """
    Look for a submodule is a series of packages, e.g. ".pagetype_plugins" in all INSTALLED_APPS.
    """
    found_apps = []
    for appconfig in apps.get_app_configs():
        app = appconfig.name
        if import_module_or_none('{0}.{1}'.format(app, submodule)) is not None:
            found_apps.append(app)

    return found_apps


def import_module_or_none(module_label):
    """
    Imports the module with the given name.

    Returns None if the module doesn't exist,
    but it does propagates import errors in deeper modules.
    """
    try:
        # On Python 3, importlib has much more functionality compared to Python 2.
        return importlib.import_module(module_label)
    except ImportError:
        # Based on code from django-oscar:
        # There are 2 reasons why there could be an ImportError:
        #
        #  1. Module does not exist. In that case, we ignore the import and return None
        #  2. Module exists but another ImportError occurred when trying to import the module.
        #     In that case, it is important to propagate the error.
        #
        # ImportError does not provide easy way to distinguish those two cases.
        # Fortunately, the traceback of the ImportError starts at __import__
        # statement. If the traceback has more than one frame, it means that
        # application was found and ImportError originates within the local app
        __, __, exc_traceback = sys.exc_info()
        frames = traceback.extract_tb(exc_traceback)
        frames = [f for f in frames
                  if f[0] != "<frozen importlib._bootstrap>" and  # Python 3.6
                  not f[0].endswith('/importlib/__init__.py')
                  and not '/_pydev_' in f[0]]
        if len(frames) > 1:
            raise
    return None



"""
Optional integration with fluent-pages features
"""
from __future__ import absolute_import

from django.utils.functional import lazy

from fluent_utils.django_compat import is_installed

__all__ = (
    'CurrentPageMixin',
    'mixed_reverse',
    'mixed_reverse_lazy',
    'HAS_APP_URLS',
)

if is_installed('fluent_pages'):
    # Use the real code.
    from fluent_pages.views import CurrentPageMixin
    from fluent_pages.urlresolvers import mixed_reverse  # app_reverse == hard dependency, no need to import here.

    try:
        from fluent_pages.urlresolvers import mixed_reverse_lazy  # added after v1.0b4
    except ImportError:
        mixed_reverse_lazy = lazy(mixed_reverse, str)

    HAS_APP_URLS = True
else:
    # Use the stubs
    from django.core.urlresolvers import reverse
    from fluent_utils.django_compat import is_installed

    try:
        from parler.views import ViewUrlMixin
    except ImportError:
        ViewUrlMixin = object

    HAS_APP_URLS = False

    class CurrentPageMixin(ViewUrlMixin):
        """
        Sub for CurrentPageMixin. Will use the real code if it exists.
        Make sure to define :attr:`view_url_name`, as that is required by :class:`~parler.views.ViewUrlMixin`.
        """
        pass

    def mixed_reverse(viewname, args=None, kwargs=None, current_app=None, current_page=None, language_code=None, multiple=False, ignore_multiple=False):
        """
        Stub for :func:`fluent_pages.urlresolvers.mixed_reverse`.
        Will use the real code if the app is loaded,
        otherwise :func:`~django.core.urlresolvers.reverse` is used.
        """
        return reverse(viewname, args=args, kwargs=kwargs, current_app=current_app)

    mixed_reverse_lazy = lazy(mixed_reverse, str)



import django
from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase

from fluent_utils.load import import_module_or_none, import_class, import_settings_class, \
    import_apps_submodule


class LoadTests(SimpleTestCase):

    def test_import_apps_submodule(self):
        expected = ['django.contrib.auth', 'django.contrib.contenttypes', 'django.contrib.sites']
        self.assertEqual(import_apps_submodule('admin'), expected)

    def test_import_settings_class(self):
        self.assertIsNotNone(import_settings_class('TEST_RUNNER'))

    def test_import_class(self):
        self.assertIs(import_class('fluent_utils.tests.test_load.LoadTests'), LoadTests)
        self.assertRaises(ImproperlyConfigured, lambda: import_class('fluent_utils.tests.test_load.InvalidLoadTests'))

    def test_ignore_top_level(self):
        self.assertIsNone(import_module_or_none("non_existant"))

    def test_detect_sub_import_error(self):
        self.assertRaises(ImportError, lambda: import_module_or_none("fluent_utils.tests.import_test"))



import fluent_utils.softdeps.any_imagefield  # noqa
import fluent_utils.softdeps.any_urlfield  # noqa
import fluent_utils.softdeps.comments  # noqa
import fluent_utils.softdeps.fluent_pages  # noqa
import fluent_utils.softdeps.taggit  # noqa



"""
Generic Ajax functionality
"""
import json

from django.http import HttpResponse


class JsonResponse(HttpResponse):
    """
    A convenient HttpResponse class, which encodes the response in JSON format.
    """

    def __init__(self, jsondata, status=200):
        self.jsondata = jsondata
        super(JsonResponse, self).__init__(json.dumps(jsondata), content_type='application/json; charset=utf-8', status=status)



try:
    from django.db.models.fields.related import ReverseOneToOneDescriptor, ForwardManyToOneDescriptor
except ImportError:
    # django < 1.9
    from django.db.models.fields.related import (
        SingleRelatedObjectDescriptor as ReverseOneToOneDescriptor,
        ReverseSingleRelatedObjectDescriptor as ForwardManyToOneDescriptor,
    )
